#pragma once

#include <algorithm>
#include <array>
#include <random>

#include "instance.hpp"

/**
 * Precalculation for frontier-based search
 * - Vertices on a frontier are labeled 0, 1, ..., |F_i|. Here F_i is the set of vertices on the i-th frontier.
 * - We distinguish between original labels  of vertices and labels of them on the frontier. The former is simply called a vertex and the latter is called an index.
 */
struct FrontierInfo {
    /**
     * @brief 
     * @param problem instance
     * @param edge_ord ordering of edges (permutation of {0, 1, ..., |E|-1})
     */
    FrontierInfo(const instance& problem, const std::vector<int>& edge_ord) : _problem(problem) {
        // reordering edges
        for (int i = 0; i < int(problem.edges.size()); ++i) {
            _problem.edges[i] = problem.edges[edge_ord[i]];
        }

        init_introduce();
        init_forget();
        init_inout();
    }

    /**
     * @brief
     * @param level 1<=level<=|E| (1: top level, |E|: bottom level)
     * @return number of vertices to introduce at the given level
     */
    int introduce_vertices_num(int level) const {
        return int(_introduce[level][0] != -1) + int(_introduce[level][1] != -1);
    }
    /**
     * @brief
     * @param level 1<=level<=|E| (1: top level, |E|: bottom level)
     * @return indices of endpoints of the introduced edge at the `level`-th frontier
     */
    std::array<int, 2> edge(int level) const {
        return _in[level];
    }
    /**
     * @brief
     * @param level 1<=level<=|E| (1: top level, |E|: bottom level)
     * @return indices of vertices to forget at the `level`-th frontier
     */
    std::array<int, 2> forget_vertices(int level) const {
        return _out[level];
    }

    /**
     * @brief 
     * @param level 1<=level<=|E| (1: top level, |E|: bottom level)
     * @return vertices on the `level`-th frontier
     */
    const std::vector<int>& frontier(int level) const {
        return _frontier_raw[level];
    }

private:
    instance _problem;
    // vertices on each frontier
    std::vector<std::vector<int>> _frontier_raw;
    // vertices to introduce for each frontier (`-1` means there are no more vertices to introduce)
    std::vector<std::array<int, 2>> _introduce;
    // vertices to forget for each frontier (`-1` means there are no more vertices to forget)
    std::vector<std::array<int, 2>> _forget;
    // indices of endpoints of the introduced edge for each frontier
    std::vector<std::array<int, 2>> _in;
    // indices to forget for each frontier (`-1` means there are no more indices to forget)
    std::vector<std::array<int, 2>> _out;

    // calculates vertices to introduce for each level
    void init_introduce() {
        const int n = _problem.n, m = _problem.edges.size();

        _introduce.assign(m + 1, { -1, -1 });
        std::vector<int8_t> seen(n, false);
        for (int i = 1; i <= m; ++i) {
            auto it = _introduce[i].begin();
            const auto& [u, v] = _problem.edges[i - 1];
            // candidates to introduce: {u, v}
            for (int x : { u, v }) if (not std::exchange(seen[x], true)) {
                *it++ = x;
            }
        }
    }
    // calculates vertices to forget for each level
    void init_forget() {
        const int n = _problem.n, m = _problem.edges.size();

        _forget.assign(m + 1, { -1, -1 });
        std::vector<int8_t> seen(n, false);
        for (int i = m; i >= 1; --i) { // traverse edges in reversed order
            auto it = _forget[i].begin();
            const auto& [u, v] = _problem.edges[i - 1];
            for (int x : { u, v }) if (not std::exchange(seen[x], true)) {
                *it++ = x;
            }
        }
    }

    /**
     * 1. calculates indices of endpoints of the introduced edge for each level
     * 2. calculates indices to forget for each level
    */
    void init_inout() {
        const int m = _problem.edges.size();

        _in.assign(m + 1, { -1, -1 });
        _out.assign(m + 1, { -1, -1 });

        _frontier_raw.resize(m + 1);
        for (int i = 1; i <= m; ++i) {
            for (int x : _introduce[i]) if (x != -1) _frontier_raw[i].push_back(x);
            auto [u, v] = _problem.edges[i - 1];

            // get indices of the endpoints of the i-th edge
            _in[i][0] = std::find(_frontier_raw[i].begin(), _frontier_raw[i].end(), u) - _frontier_raw[i].begin();
            _in[i][1] = std::find(_frontier_raw[i].begin(), _frontier_raw[i].end(), v) - _frontier_raw[i].begin();

            int siz = _frontier_raw[i].size();

            // copy to the next frontier and will remove some vertices from it
            if (i != m) _frontier_raw[i + 1] = _frontier_raw[i];

            for (int j = siz - 1; j >= 0; --j) {
                int x = _frontier_raw[i][j];
                if (x == _forget[i][0]) { // found a vertex to forget
                    _out[i][0] = j;
                    // remove x from the next frontier
                    if (i != m) _frontier_raw[i + 1].erase(_frontier_raw[i + 1].begin() + j);
                } else if (x == _forget[i][1]) { // found a vertex to forget
                    _out[i][1] = j;
                    // remove x from the next frontier
                    if (i != m) _frontier_raw[i + 1].erase(_frontier_raw[i + 1].begin() + j);
                }
            }
        }
    }
};

/**
 * @brief Configuration for frontier-based search
 */
struct FrontierState {
private:
    /**
     * @brief store the precalculated information
     */
    struct LevelInfo {
        int introduced_vertex_num; // number of vertices to introduce
        int edge_src; // an index of one endpoint of the introduced edge
        int edge_dst; // an index of the other endpoint of the introduced edge
        std::array<int, 2> forget_index; // indices to forget

        std::vector<int> frontier; // vertices on the current frontier

        LevelInfo() = default;

        /**
         * @brief update frontier information
         * @param level new level
         * @param info precalculated information
         */
        void set_level(int level, const FrontierInfo& info) {
            introduced_vertex_num = info.introduce_vertices_num(level);
            edge_src = info.edge(level)[0];
            edge_dst = info.edge(level)[1];
            forget_index = info.forget_vertices(level);
            frontier = info.frontier(level);
        }
    };

    // Zobrist Hash
    struct PathHash {
        PathHash() = default;
        PathHash(const instance& problem) : n(problem.get_n()), h((n + 1)* (n + 1)) {
            // assignment hashes randomly
            std::mt19937_64 rng{};
            for (size_t i = 0; i <= n; ++i) for (size_t j = 0; j <= i; ++j) {
                h[i * (n + 1) + j] = h[j * (n + 1) + i] = rng();
            }
        }
        /**
         * @brief
         * @param i index of one endpoint (on the current frontier) of the path
         * @param j index of the other endpoint (on the current frontier) of the path
         * @return hash of path(i, j)
         */
        uint64_t path_hash(size_t i, size_t j) const {
            return h[i * (n + 1) + j];
        }
        /**
         * @brief 
         * @param i index of a vertex with degree 2
         * @return hash of the index i with degree 2
         */
        uint64_t deg2_hash(size_t i) const {
            return h[i * (n + 1) + i];
        }
        /**
         * @brief 
         * @param i index of one endpoint (on the current frontier) of the path
         * @return hash of path(i, j) (j is the other endpoint forgotten from the frontier)
         */
        uint64_t path_hash_forgotten(size_t i) const {
            return h[n * (n + 1) + i];
        }
    private:
        size_t n;
        std::vector<uint64_t> h;
    } static inline _path_hasher{};

public:
    FrontierState() = default;

    /**
     * @brief call this method to specify terminals (only for one-pair instance)
     * @param s one terminal (0<=s<|V|)
     * @param t the other terminal (0<=t<|V|)
     */
    static void set_terminals(int s, int t) {
        _s = s;
        _t = t;
    }
    /**
     * @brief update frontier information
     * @param level new level
     * @param info precalculated information
     */
    static void set_level(int level, const FrontierInfo& info) {
        _info.set_level(level, info);
        _is_terminal_src = _info.frontier[_info.edge_src] == _s or _info.frontier[_info.edge_src] == _t;
        _is_terminal_dst = _info.frontier[_info.edge_dst] == _s or _info.frontier[_info.edge_dst] == _t;
    }
    /**
     * @brief check if the index i is one of the terminals
     * @param i index (0<=i<|F|, where F is the set of vertices on the current frontier)
     * @return true if the index i is one of the terminals, false otherwise
     */
    static bool is_terminal(int i) {
        return _info.frontier[i] == _s or _info.frontier[i] == _t;
    }

    /**
     * @brief call this before running frontier-based search
     * @param problem instance
     */
    static void init_hasher(const instance& problem) {
        _path_hasher = PathHash(problem);
    }

    /**
     * @brief representing that there is no mate (degree: 0)
     */
    static constexpr int no_mate = -2;
    /**
     * @brief representing that there is a forgotten mate (degree: 1)
     */
    static constexpr int forgotten_mate = -3;
    /**
     * @brief representing an intermediate point (degree: 2)
     */
    static constexpr int deg2 = -4;

    enum class TransitionResult {
        ONWAY,
        FAIL,
        COMPLETE,
    };

    /**
     * @brief Check for violations of some conditions that can be easily determined.
     * @tparam check_terminal whether checking the conditions about endpoints of paths or not
     * @return true if no violations are detected, false otherwise
     */
    template <bool check_terminal>
    bool roughly_check_if_new_edge_can_be_added() const {
        const int i = _info.edge_src, j = _info.edge_dst;
        const int mate_i = i < int(_mate.size()) ? _mate[i] : no_mate;
        const int mate_j = j < int(_mate.size()) ? _mate[j] : no_mate;

        if constexpr (check_terminal) {
            if (_is_terminal_src) {
                // The degrees of the terminal vertices cannot be exceed 1
                if (mate_i != no_mate) return false;
            } else {
                // The degrees of the other vertices cannot be exceed 2
                if (mate_i == deg2) return false;
            }
            if (_is_terminal_dst) {
                if (mate_j != no_mate) return false;
            } else {
                if (mate_j == deg2) return false;
            }
        } else {
            // The degrees of vertices cannot be exceed 2
            if (mate_i == deg2) return false;
            if (mate_j == deg2) return false;
        }
        // There must not be any cycles.
        if (mate_i == j) {
            return false;
        }
        return true;
    }

    /**
     * @brief perform inplace state transitions
     * @tparam check_terminal whether checking the conditions about endpoints of paths or not
     * @param use_edge whether using the edge at the current level or not
     * @param roughly_checked whether you have already checked for some condition violations by calling `roughly_check_if_new_edge_can_be_added` or not
     * @return TransitionResult::ONWAY, TransitionResult::COMPLETE, or TransitionResult::FAIL
     */
    template <bool check_terminal>
    TransitionResult make_transition(bool use_edge, bool roughly_checked = false) {
        // introduce vertices
        for (int i = 0; i < _info.introduced_vertex_num; ++i) {
            _mate.push_back(no_mate);
        }

        if (use_edge) {
            const int i = _info.edge_src, j = _info.edge_dst;

            if (not roughly_checked) {
                if (not roughly_check_if_new_edge_can_be_added<check_terminal>()) {
                    return TransitionResult::FAIL;
                }
            }

            if (_mate[i] == no_mate) {
                if (_mate[j] == no_mate) {
                    // i doesn't have a mate / j doesn't have a mate

                    _mate[i] = j;
                    _mate[j] = i;
                    ++_path_num;

                    const int vi = _info.frontier[i], vj = _info.frontier[j];
                    // add path(i, j)
                    _hash ^= _path_hasher.path_hash(vi, vj);
                } else if (_mate[j] == forgotten_mate) {
                    // i doesn't have a mate / j has a forgotten mate

                    _mate[i] = forgotten_mate;
                    _mate[j] = deg2;

                    const int vi = _info.frontier[i], vj = _info.frontier[j];
                    // rem path(j, ?)
                    _hash ^= _path_hasher.path_hash_forgotten(vj);
                    // add deg2(j)
                    _hash ^= _path_hasher.deg2_hash(vj);
                    // add path(i, ?)
                    _hash ^= _path_hasher.path_hash_forgotten(vi);
                } else {
                    // i doesn't have a mate / j has a mate k on the current frontier

                    int k = _mate[j];
                    _mate[i] = k;
                    _mate[j] = deg2;
                    _mate[k] = i;

                    const int vi = _info.frontier[i], vj = _info.frontier[j], vk = _info.frontier[k];
                    // rem path(j, k)
                    _hash ^= _path_hasher.path_hash(vk, vj);
                    // add path(i, k)
                    _hash ^= _path_hasher.path_hash(vk, vi);
                    // add deg2(j)
                    _hash ^= _path_hasher.deg2_hash(vj);
                }
            } else if (_mate[i] == forgotten_mate) {
                if (_mate[j] == no_mate) {
                    // i has a forgotten mate / j doesn't have a mate

                    _mate[i] = deg2;
                    _mate[j] = forgotten_mate;

                    const int vi = _info.frontier[i], vj = _info.frontier[j];
                    // rem path(i, ?)
                    _hash ^= _path_hasher.path_hash_forgotten(vi);
                    // add path(j, ?)
                    _hash ^= _path_hasher.path_hash_forgotten(vj);
                    // add deg2(i)
                    _hash ^= _path_hasher.deg2_hash(vi);
                } else if (_mate[j] == forgotten_mate) {
                    // i has a forgotten mate / j has a forgotten mate

                    _mate[i] = deg2;
                    _mate[j] = deg2;
                    // It's ok not to update hash since this state will not be referenced any more.
                    return --_path_num == 1 ? TransitionResult::COMPLETE : TransitionResult::FAIL;
                } else {
                    // i has a forgotten mate / j has a mate k on the current frontier

                    int k = _mate[j];
                    _mate[i] = deg2;
                    _mate[j] = deg2;
                    _mate[k] = forgotten_mate;
                    --_path_num;

                    const int vi = _info.frontier[i], vj = _info.frontier[j], vk = _info.frontier[k];
                    // rem path(j, k)
                    _hash ^= _path_hasher.path_hash(vj, vk);
                    // rem path(i, ?)
                    _hash ^= _path_hasher.path_hash_forgotten(vi);
                    // add path(k, ?)
                    _hash ^= _path_hasher.path_hash_forgotten(vk);
                    // add deg2(i)
                    _hash ^= _path_hasher.deg2_hash(vi);
                    // add deg2(j)
                    _hash ^= _path_hasher.deg2_hash(vj);
                }
            } else {
                if (_mate[j] == no_mate) {
                    // i has a mate k on the current frontier / j doesn't have a mate

                    int k = _mate[i];
                    _mate[i] = deg2;
                    _mate[j] = k;
                    _mate[k] = j;

                    const int vi = _info.frontier[i], vj = _info.frontier[j], vk = _info.frontier[k];
                    // rem path(i, k)
                    _hash ^= _path_hasher.path_hash(vk, vi);
                    // add path(j, k)
                    _hash ^= _path_hasher.path_hash(vk, vj);
                    // add deg2(i)
                    _hash ^= _path_hasher.deg2_hash(vi);
                } else if (_mate[j] == forgotten_mate) {
                    // i has a mate k on the current frontier / j has a forgotten mate

                    int k = _mate[i];
                    _mate[i] = deg2;
                    _mate[j] = deg2;
                    _mate[k] = forgotten_mate;
                    --_path_num;

                    const int vi = _info.frontier[i], vj = _info.frontier[j], vk = _info.frontier[k];
                    // rem path(i, k)
                    _hash ^= _path_hasher.path_hash(vi, vk);
                    // rem path(j, ?)
                    _hash ^= _path_hasher.path_hash_forgotten(vj);
                    // add path(k, ?)
                    _hash ^= _path_hasher.path_hash_forgotten(vk);
                    // add deg2(i)
                    _hash ^= _path_hasher.deg2_hash(vi);
                    // add deg2(j)
                    _hash ^= _path_hasher.deg2_hash(vj);
                } else {
                    // i has a mate k on the current frontier / j has a mate l on the current frontier

                    int k = _mate[i];
                    int l = _mate[j];
                    _mate[i] = deg2;
                    _mate[j] = deg2;
                    _mate[k] = l;
                    _mate[l] = k;
                    --_path_num;

                    const int vi = _info.frontier[i], vj = _info.frontier[j], vk = _info.frontier[k], vl = _info.frontier[l];
                    // rem path(i, k)
                    _hash ^= _path_hasher.path_hash(vi, vk);
                    // rem path(j, l)
                    _hash ^= _path_hasher.path_hash(vj, vl);
                    // add path(k, l)
                    _hash ^= _path_hasher.path_hash(vk, vl);
                    // add deg2(i)
                    _hash ^= _path_hasher.deg2_hash(vi);
                    // add deg2(j)
                    _hash ^= _path_hasher.deg2_hash(vj);
                }
            }
        }

        for (int i : _info.forget_index) if (i != -1) {
            // forget a vertex with degree 0
            if (_mate[i] == no_mate) {
                if constexpr (check_terminal) {
                    // terminals must have degree 1, not 0
                    if (is_terminal(i)) return TransitionResult::FAIL;
                }
                continue;
            }

            // forget a vertex with degree 1
            // it will be one of the endpoints of a path
            if (_mate[i] != deg2) {
                if constexpr (check_terminal) {
                    // i must be match with one of the vertices specified as a terminal
                    if (not is_terminal(i)) {
                        return TransitionResult::FAIL;
                    }
                }
                // number of endpoints must be less than or equal to 2
                if (++_endpoint >= 3) {
                    return TransitionResult::FAIL;
                }
            }

            if (int k = _mate[i]; k >= 0) {
                _mate[k] = forgotten_mate;

                // rem path(k, i)
                _hash ^= _path_hasher.path_hash(_info.frontier[k], _info.frontier[i]);
                // add path(k, ?)
                _hash ^= _path_hasher.path_hash_forgotten(_info.frontier[k]);
            } else if (k == forgotten_mate) {
                // It's ok not to update hash since this state will not be referenced any more.
                return _path_num == 1 ? TransitionResult::COMPLETE : TransitionResult::FAIL;
            } else if (k == deg2) {
                // rem deg2(k)
                _hash ^= _path_hasher.deg2_hash(_info.frontier[i]);
            }
        }

        auto [fi, fj] = _info.forget_index;
        if (fi > fj) std::swap(fi, fj);
        for (int i : { fj, fi }) {
            if (i == -1) break;
            // adjusting indices
            for (auto& m : _mate) {
                m -= (m > i);
            }
            _mate.erase(_mate.begin() + i);
        }
        return TransitionResult::ONWAY;
    }

    /**
     * @brief 
     * @return hash of the current state
     */
    std::size_t hash() const {
        // size_t res = 0;
        // for (int e : _mate) res = res * 31 + e;
        // return res;
        return _hash;
    }
    friend bool operator==(const FrontierState& l, const FrontierState& r) {
        return l._mate == r._mate;
    }

    /**
     * @brief 
     * @return number of paths with positive length
     */
    int path_num() const {
        return _path_num;
    }

    static const LevelInfo& frontier_info() {
        return _info;
    }
private:
    // storing the precalculated information about the frontier at the current level
    static inline LevelInfo _info;
    // specified terminals (only for one-pair instances)
    static inline int _s = -1, _t = -1;
    // whether endpoints of the edge to introduce are specified terminals or not
    static inline bool _is_terminal_src = false, _is_terminal_dst = false;

    // so called mate array used in Simpath algorithm
    std::vector<int8_t> _mate{};
    
    // --- auxiliary information --- //

    // the hash value of this state
    size_t _hash = 0;
    // the number of forgotten vertices with degree 1
    uint8_t _endpoint = 0;
    // the number of paths with positive length
    uint8_t _path_num = 0;
};

template <>
struct std::hash<FrontierState> {
    std::size_t operator() (const FrontierState& key) const {
        return key.hash();
    }
};