#pragma once

#include <cassert>
#include <cstdint>
#include <iostream>
#include <deque>
#include <limits>
#include <numeric>
#include <set>
#include <sstream>
#include <vector>

struct instance {
    // number of vertices (vertices are labeled 0,1,...,|V|-1)
    int n;
    // set of edges
    std::vector<std::pair<int, int>> edges;
    // terminals (-1 if not specified)
    int s, t;
    // maximum path length
    int l;

    instance() = default;
    instance(int n, const std::vector<std::pair<int, int>>& edges, int s, int t, int l) : n(n), edges(edges), s(s), t(t), l(l) {}

    instance(const std::string& input_str) {
        std::istringstream iss{ input_str };
        read_DIMACS(iss);
    }

    int get_n() const { return n; }
    int get_s() const { return s; }
    int get_t() const { return t; }
    int get_l() const { return l; }
    const std::vector<std::pair<int, int>>& get_edges() const { return edges; }
    std::vector<std::pair<int, int>>& get_edges() { return edges; }

    /**
     * @brief construct adjacency list
     * @return adjacency list
     */
    std::vector<std::vector<int>> adjacent_list() const {
        std::vector<std::vector<int>> adj(n);
        for (auto [u, v] : edges) {
            adj[u].push_back(v);
            if (u != v) adj[v].push_back(u);
        }
        return adj;
    }

    /**
     * @brief 
     * @return an instance with V={0}, E={}, s=t=0, l=0
     */
    static instance trivial_instance() {
        return { 1, {}, 0, 0, 0 };
    }
    /**
     * @brief 
     * @return an instance with V={}, E={}, s=t=0, l=0
     */
    static instance invalid_instance() {
        return { 0, {}, 0, 0, 0 };
    }
    bool operator==(const instance& other) const {
        return n == other.n and edges == other.edges and s == other.s and t == other.t and l == other.l;
    }
    bool is_invalid() const {
        return *this == invalid_instance();
    }
    bool is_trivial() const {
        return *this == trivial_instance();
    }

    /**
     * @brief create a new instance by removing a single vertex.
     * @param v a vertex to remove
     * @return a new instance created by removing a single vertex
     */
    instance erase_vertex(int v) const {
        return erase_vertices({ v });
    }
    /**
     * @brief create a new instance by removing vertices.
     * @param v vertices to remove
     * @return a new instance created by removing vertices
     */
    instance erase_vertices(const std::vector<int>& vs) const {
        if (is_invalid()) return invalid_instance();

        instance res;

        std::vector<int8_t> is_erased(n);
        for (int v : vs) {
            is_erased[v] = true;
        }

        if (is_erased[s] or is_erased[t]) {
            return invalid_instance();
        }

        res.n = n - vs.size();

        // compute new labels for remaining vertices
        std::vector<int> id(n, -1);
        for (int i = 0, nxt_id = 0; i < n; ++i) {
            if (not is_erased[i]) {
                id[i] = nxt_id++;
            }
        }

        for (auto [u, v] : edges) {
            if (is_erased[u] or is_erased[v]) continue;
            res.edges.emplace_back(id[u], id[v]);
        }
        res.s = id[s], res.t = id[t];
        res.l = l;

        return res;
    }

    /**
     * @brief equivalent to erase_vertices(V-vs)
     */
    instance induced(const std::vector<int>& vs) const {
        std::vector<int8_t> is_erased(n, true);
        for (int v : vs) is_erased[v] = false;
        std::vector<int> erase_vs;
        for (int i = 0; i < n; ++i) if (is_erased[i]) erase_vs.push_back(i);
        return erase_vertices(erase_vs);
    }

    /**
     * @brief reduce the instance by (1) removing vertices not included in any of s-t simple paths with length l or less and by (2) removing vertices with degree 1.
     * @return reduced instance
     */
    instance reduce() const {
        if (is_invalid()) return invalid_instance();

        std::vector<std::vector<int>> adj = adjacent_list();

        constexpr int inf = std::numeric_limits<int>::max() / 2;

        // compute distance from the vertex `start` to all vertices
        auto bfs = [&](int start) {
            std::vector<int> dist(n, inf);
            dist[start] = 0;

            std::deque<int> dq{ start };
            while (dq.size()) {
                int u = dq.front();
                dq.pop_front();
                for (int v : adj[u]) if (dist[v] == inf) {
                    dist[v] = dist[u] + 1;
                    dq.push_back(v);
                }
            }
            return dist;
            };

        // removing vertices not included in any of s-t simple paths with length l or less
        {
            std::vector<int> ds = bfs(s);
            if (ds[t] > l) return invalid_instance();
            std::vector<int> dt = bfs(t);
            std::vector<int> erase_vs;
            for (int i = 0; i < n; ++i) {
                if (ds[i] + dt[i] > l) erase_vs.push_back(i);
            }
            if (not erase_vs.empty()) {
                return erase_vertices(erase_vs).reduce();
            }
        }

        std::vector<int> deg(n);
        for (int i = 0; i < n; ++i) {
            deg[i] = adj[i].size();
        }
        std::vector<int8_t> removed(n);

        instance reduced_problem = *this;

        auto remove = [&](int v) {
            removed[v] = true;
            for (int w : adj[v]) if (not removed[w]) {
                if (v == reduced_problem.s) {
                    // s has degree 1
                    reduced_problem.s = w;
                    --reduced_problem.l;
                } else if (v == reduced_problem.t) {
                    // t has degree 1
                    reduced_problem.t = w;
                    --reduced_problem.l;
                }
                --deg[w];
            }
            };

        // removing vertices with degree 1
        std::deque<int> dq;
        for (int v = 0; v < n; ++v) if (deg[v] == 1) {
            dq.push_back(v);
        }

        while (dq.size()) {
            if (reduced_problem.s == reduced_problem.t) return trivial_instance();

            int u = dq.front();
            dq.pop_front();
            remove(u);

            for (int v : adj[u]) if (not removed[v] and deg[v] == 1) {
                dq.push_back(v);
            }
        }
        if (reduced_problem.s == reduced_problem.t) return trivial_instance();

        assert(reduced_problem.l >= 0);

        std::vector<int> erase_vs;
        for (int i = 0; i < n; ++i) if (removed[i]) {
            erase_vs.push_back(i);
        }
        return reduced_problem.erase_vertices(erase_vs);
    }

    /**
     * @brief check if this instance has already been reduced or not
     * @return true if reduced, false otherwise
     */
    bool is_reduced() const {
        return reduce() == *this;
    }

    /**
     * @brief read an instance from istream
     * @param in istream (std::cin if you want to read from stdin)
     */
    void read_DIMACS(std::istream& in) {
        int m = -1;
        n = s = t = l = -1;
        edges.clear();

        std::string line;
        std::string ignore; // for wasting input
        while (std::getline(in, line)) {
            if (line.empty()) continue;
            std::istringstream iss{ line };
            char op;
            iss >> op;
            if (op == 'c') {
                continue;
            } else if (op == 'p') {
                iss >> ignore >> n >> m;
                assert(ignore == "edge");
            } else if (op == 'e') {
                int u, v;
                iss >> u >> v;
                --u, --v;
                edges.emplace_back(u, v);
            } else if (op == 'l') {
                iss >> l;
            } else if (op == 't') {
                iss >> s >> t;
                --s, --t;
            } else {
                assert(false);
            }
        }
        assert(n >= 0);
        // assert(s >= 0 and t >= 0);
        assert(l >= 0);
        assert(m == int(edges.size()));
    }

    void write_info(std::ostream& out) {
        out << "V = " << n << std::endl;
        out << "E = " << edges.size() << std::endl;
    }
    void write_DIMACS(std::ostream& out) {
        out << "p edge " << n << ' ' << edges.size() << '\n';
        for (auto [u, v] : edges) {
            out << "e " << u + 1 << ' ' << v + 1 << '\n';
        }
        out << "l " << l << '\n';
        if (s != -1) {
            out << "t " << s + 1 << ' ' << t + 1 << std::endl;
        }
    }

    void write_cp_format(std::ostream& out) {
        out << n << ' ' << edges.size() << '\n';
        for (auto [u, v] : edges) {
            out << u + 1 << ' ' << v + 1 << '\n';
        }
    }

    /**
     * @brief equivalence partitioning by the equivalence relation u~v <==def==> N(u)-{v}=N(v)-{u}
     * @return pair { number of equivalence classes, equivalence classes }
     */
    std::pair<int, std::vector<int>> neighborhood_division() const {
        const auto adj = adjacent_list();

        // check if i~j or not
        auto same_neighborhood = [&](int i, int j) {
            std::vector<int> adj_i = adj[i];
            std::vector<int> adj_j = adj[j];
            if (auto it = std::find(adj_i.begin(), adj_i.end(), j); it != adj_i.end()) {
                adj_i.push_back(i);
                adj_j.push_back(j);
            }
            std::sort(adj_i.begin(), adj_i.end());
            std::sort(adj_j.begin(), adj_j.end());
            return adj_i == adj_j;
            };

        std::vector<int> group_id(n);
        std::iota(group_id.begin(), group_id.end(), 0);
        for (int i = 0; i < n; ++i) for (int j = i + 1; j < n; ++j) {
            if (group_id[i] == group_id[j] or not same_neighborhood(i, j)) continue;
            for (auto& e : group_id) if (e == group_id[j]) {
                e = group_id[i];
            }
        }
        std::vector<int> ids = group_id;
        std::sort(ids.begin(), ids.end());
        ids.erase(std::unique(ids.begin(), ids.end()), ids.end());
        for (auto& e : group_id) {
            e = std::lower_bound(ids.begin(), ids.end(), e) - ids.begin();
        }
        const int group_num = ids.size();
        return { group_num, group_id };
    }

    /**
     * @brief reduce to weighted instance using neighborhood division (each new vertex represents one equavalence class)
     * @return pair { instance, weights of vertices } (weights mean the cardinalities of equivalence classes)
     */
    std::pair<instance, std::vector<int>> reduce_neighborhood_division() const {
        auto [group_num, group_id] = neighborhood_division();

        std::vector<int> weight(group_num);
        for (auto& e : group_id) {
            ++weight[e];
        }
        std::set<std::pair<int, int>> new_edges;
        for (auto [u, v] : edges) {
            new_edges.insert(std::minmax(group_id[u], group_id[v]));
        }
        instance new_problem(
            group_num,
            std::vector<std::pair<int, int>>(new_edges.begin(), new_edges.end()),
            s != -1 ? group_id[s] : s,
            t != -1 ? group_id[t] : t,
            l
        );
        return { new_problem, weight };
    }
};
