#pragma once

#include <algorithm>
#include <cstdint>
#include <vector>

#include <omp.h>

#include "instance.hpp"
#include "graph/bfs.hpp"

#include "util/robin_hood.h"

namespace detail {
    template <typename T>
    struct SortedSequence {
    private:
        // zobrist hashing
        static inline std::vector<size_t> _hasher{};

        size_t _hash = 0;

        // sequence
        std::vector<T> _dat{};
    public:
        SortedSequence() = default;
        template <typename Container>
        SortedSequence(const Container &dat) {
            for (const T &x : dat) add(x);
        }
        static void init_hash(size_t n) {
            std::mt19937_64 rng{};
            _hasher.resize(n);
            std::generate(_hasher.begin(), _hasher.end(), rng);
        }
        size_t size() const {
            return _dat.size();
        }
        void reserve(size_t n) {
            _dat.reserve(n);
        }
        /**
         * @brief Add an element to the sequence. **Elements must be added in ascending order.**
         * @param x an element to add
         */
        void add(const T &x) {
            _hash ^= _hasher[x];
            assert(_dat.empty() or _dat.back() <= x);
            _dat.push_back(x);
        }
        /**
         * @brief Remove one of the maximum elements
         */
        void pop() {
            _hash ^= _hasher[_dat.back()];
            _dat.pop_back();
        }
        void clear() {
            _hash = 0;
            _dat.clear();
        }
        size_t hash() const {
            return _hash;
        }
        friend bool operator==(const SortedSequence &s, const SortedSequence &t) {
            return s._dat == t._dat;
        }
    };
}

template <typename T>
struct std::hash<detail::SortedSequence<T>> {
    size_t operator()(const detail::SortedSequence<T>& s) const {
        return s.hash();
    }
};

namespace detail {
    /**
     * @brief solve one-pair instances using meet-in-the-middle technique 
     * @tparam Int integer type used to count
     * @param problem one-pair instance
     * @return number of paths grouped by length
     */
    template <typename Int>
    std::vector<Int> solve_dfs_meet_in_the_middle_one_pair(const instance& problem) {
        if (problem.is_invalid()) {
            return std::vector<Int>{ 0 };
        } else if (problem.is_trivial()) {
            return std::vector<Int>{ 1 };
        }

        const int n = problem.n;

        // maximum m-s path length (m: midpoint of paths)
        const int ms_maxlen = (problem.get_l() + 1) / 2;
        // maximum m-t path length (m: midpoint of paths)
        const int mt_maxlen = problem.get_l() / 2;

        const auto adj = problem.adjacent_list();

        const auto ds = bfs(adj, problem.s);
        const auto dt = bfs(adj, problem.t);

        const auto [group_num, group_id] = problem.neighborhood_division();
        std::vector<std::vector<int>> groups(group_num);
        for (int i = 0; i < n; ++i) {
            groups[group_id[i]].push_back(i);
        }

        // computing answers for each equivalent class is enough
        std::vector<std::vector<Int>> answers(group_num, std::vector<Int>(problem.get_l() + 1));

        using Set = SortedSequence<int>;
        Set::init_hash(n);
        using HashMap = robin_hood::unordered_map<Set, std::vector<Int>>;

        std::cerr << "Neighborhood Diversity = " << group_num << std::endl;

#pragma omp parallel for schedule(dynamic,1) num_threads(16)
        for (int id = 0; id < group_num; ++id) {
            // midpoint
            const int m = [&]{
                for (int m : groups[id]) {
                    if (m == problem.s) continue;
                    if (m == problem.t) continue;
                    return m;
                }
                return -1;
            }();
            if (m >= 0 and ds[m] <= ms_maxlen and dt[m] <= mt_maxlen) {
                std::vector<Int>& ans = answers[id];
                HashMap cnt;

                // add all subsets of V(ms_path) to `cnt`
                const auto insert_all_subsets = [&](const std::vector<int> &ms_path){
                    const int len = ms_path.size();
                    std::vector<int> sorted_ms_path = ms_path;
                    sorted_ms_path.pop_back(); // remove s

                    // pre-sorting
                    std::sort(sorted_ms_path.begin(), sorted_ms_path.end());

                    Set subset;
                    subset.reserve(len);
                    for (int bits = 0; bits < 1 << (len - 1); ++bits) {
                        subset.clear();
                        // elements are added in ascending order
                        for (int i = 0; i < len - 1; ++i) if ((bits >> i) & 1) {
                            subset.add(sorted_ms_path[i]);
                        }
                        auto& c = cnt[subset];
                        if (int(c.size()) <= len) {
                            c.resize(len + 1);
                        }
                        ++c[len];
                    }
                };
                // counting s-t simple paths using the inclusion-exclusion principle
                const auto inclusion_exclusion = [&](const std::vector<int> &mt_path) {
                    const int len = mt_path.size();
                    std::vector<int> sorted_mt_path = mt_path;
                    sorted_mt_path.pop_back(); // remove t
                    std::sort(sorted_mt_path.begin(), sorted_mt_path.end());

                    Set subset;
                    // Here we run dfs for enumerating subsets since the number of valid subsets is expected to be small.
                    auto all_subsets = [&](auto all_subsets, int nxt_begin) -> void {
                        auto it = cnt.find(subset);
                        // pruning
                        if (it == cnt.end()) return;

                        const auto& c = it->second;
                        Int sgn = (subset.size() & 1) ? -1 : +1;
                        if (int(c.size()) > len) {
                            ans[len + len] += sgn * c[len];
                            if (int(c.size()) > len + 1) {
                                ans[len + len + 1] += sgn * c[len + 1];
                            }
                        }

                        // enumerating larger subsets by recursive calling
                        for (int nxt = nxt_begin; nxt < len - 1; ++nxt) {
                            subset.add(sorted_mt_path[nxt]);
                            all_subsets(all_subsets, nxt + 1);
                            subset.pop();
                        }
                    };
                    all_subsets(all_subsets, 0);
                };

                std::vector<int8_t> visited(n);
                visited[m] = true;

                std::vector<int> ms_path;
                // run dfs for enumerating m-s paths
                auto enumerate_ms_paths = [&](auto rec, int u) -> void {
                    const int len = ms_path.size();
                    if (u == problem.s) {
                        insert_all_subsets(ms_path);
                    } else {
                        for (int v : adj[u]) if (not visited[v] and len + 1 + ds[v] <= ms_maxlen and v != problem.t) {
                            visited[v] = true;
                            ms_path.push_back(v);
                            rec(rec, v);
                            ms_path.pop_back();
                            visited[v] = false;
                        }
                    }
                };
                enumerate_ms_paths(enumerate_ms_paths, m);

                std::vector<int> mt_path;
                // run dfs for enumerating m-t paths
                auto enumerate_mt_paths = [&](auto rec, int u) -> void {
                    const int len = mt_path.size();
                    if (u == problem.t) {
                        inclusion_exclusion(mt_path);
                    } else {
                        for (int v : adj[u]) if (not visited[v] and len + 1 + dt[v] <= mt_maxlen and v != problem.s) {
                            visited[v] = true;
                            mt_path.push_back(v);
                            rec(rec, v);
                            mt_path.pop_back();
                            visited[v] = false;
                        }
                    }
                };
                enumerate_mt_paths(enumerate_mt_paths, m);
            }

            std::string msg = "Done " + std::to_string(id) + "\n";
            std::cerr << msg;
            std::cerr.flush();
        }
        std::vector<Int> ans(problem.get_l() + 1);
        for (int v = 0; v < n; ++v) if (v != problem.s and v != problem.t) {
            for (int i = 1; i <= problem.get_l(); ++i) {
                ans[i] += answers[group_id[v]][i];
            }
        }
        // paths of length 1 are not yet counted
        if (problem.get_l() >= 1 and std::find(adj[problem.s].begin(), adj[problem.s].end(), problem.t) != adj[problem.s].end()) {
            ++ans[1];
        }
        return ans;
    }

    /**
     * @brief solve all-pairs instances using meet-in-the-middle technique 
     * @tparam Int integer type used to count
     * @param problem all-pairs instance
     * @return number of paths grouped by length
     */
    template <typename Int>
    std::vector<Int> solve_dfs_meet_in_the_middle_all_pairs(const instance& problem) {
        if (problem.is_invalid()) {
            return std::vector<Int>{ 0 };
        } else if (problem.is_trivial()) {
            return std::vector<Int>{ 1 };
        }

        const int n = problem.n;
        const int maxlen = (problem.get_l() + 1) / 2;

        const auto adj = problem.adjacent_list();

        const auto [group_num, group_id] = problem.neighborhood_division();
        std::vector<std::vector<int>> groups(group_num);
        for (int i = 0; i < n; ++i) {
            groups[group_id[i]].push_back(i);
        }

        // computing answers for each equivalent class is enough
        std::vector<std::vector<Int>> answers(group_num, std::vector<Int>(problem.get_l() + 3));

        using Set = SortedSequence<int>;
        Set::init_hash(n);
        using HashMap = robin_hood::unordered_map<Set, std::vector<Int>>;

        std::cerr << "Neighborhood Diversity = " << group_num << std::endl;

#pragma omp parallel for schedule(dynamic,1) num_threads(16)
        for (int id = 0; id < group_num; ++id) {
            // midpoint
            const int m = groups[id].front();

            std::vector<Int>& ans = answers[id];

            HashMap cnt;

            std::vector<int8_t> visited(n);
            visited[m] = true;

            std::vector<int> path;
            // run dfs and count subsets of paths and the answer simultaneously
            auto enumerate_paths = [&](auto rec, int u) -> void {
                const int len = path.size();
                if (len) {
                    std::vector<int> sorted_path = path;

                    // pre-sorting
                    std::sort(sorted_path.begin(), sorted_path.end());

                    Set subset;
                    subset.reserve(len);
                    for (int bits = 0; bits < 1 << len; ++bits) {
                        subset.clear();
                        // elements are added in ascending order
                        for (int i = 0; i < len; ++i) if ((bits >> i) & 1) {
                            subset.add(sorted_path[i]);
                        }
                        auto& c = cnt[subset];

                        // inclusion-exclusion
                        Int sgn = (subset.size() & 1) ? -1 : +1;
                        if (int(c.size()) > len - 1) {
                            ans[len + len - 1] += sgn * c[len - 1];
                            if (int(c.size()) > len) {
                                ans[len + len] += sgn * c[len];
                                if (int(c.size()) > len + 1) {
                                    ans[len + len + 1] += sgn * c[len + 1];
                                }
                            }
                        }

                        // insert subset of path
                        if (int(c.size()) <= len) {
                            c.resize(len + 1);
                        }
                        ++c[len];
                    }
                }
                for (int v : adj[u]) if (not visited[v] and len + 1 <= maxlen) {
                    visited[v] = true;
                    path.push_back(v);
                    rec(rec, v);
                    path.pop_back();
                    visited[v] = false;
                }
            };
            enumerate_paths(enumerate_paths, m);

            std::string msg = "Done " + std::to_string(id) + "\n";
            std::cerr << msg;
            std::cerr.flush();
        }
        std::vector<Int> ans(problem.get_l() + 1);
        for (int v = 0; v < n; ++v) {
            for (int i = 1; i <= problem.get_l(); ++i) {
                ans[i] += answers[group_id[v]][i];
            }
        }
        if (problem.get_l() >= 1) {
            ans[1] += problem.edges.size();
        }
        // paths with odd lengths are counted exactly twice
        for (int i = 3; i <= problem.get_l(); i += 2) {
            ans[i] /= 2;
        }
        return ans;
    }
}

template <typename Int>
std::vector<Int> solve_dfs_meet_in_the_middle(const instance& problem) {
    if (problem.is_invalid()) {
        return std::vector<Int>{ 0 };
    } else if (problem.is_trivial()) {
        return std::vector<Int>{ 1 };
    }
    if (problem.s == -1) {
        return detail::solve_dfs_meet_in_the_middle_all_pairs<Int>(problem);
    } else {
        return detail::solve_dfs_meet_in_the_middle_one_pair<Int>(problem);
    }
}
