#pragma once

#include <algorithm>
#include <iostream>
#include <iomanip>
#include <vector>

#include "instance.hpp"
#include "frontier.hpp"

#include "util/progress_bar.hpp"

#include "util/robin_hood.h"

#include "omp.h"

namespace detail {
    /**
     * @brief solve instances with a no limitation of path length by Simpath algorithm
     * @tparam Int integer type used to count
     * @tparam check_terminal whether checking the conditions about endpoints of paths or not
     * @param problem instance
     * @param edge_ord ordering of edges
     * @return number of paths grouped by length
     */
    template <typename Int, bool check_terminal>
    Int solve_frontier(const instance& problem, const std::vector<int>& edge_ord) {
        const FrontierInfo info(problem, edge_ord);

        FrontierState::set_terminals(problem.s, problem.t);
        FrontierState::init_hasher(problem);

        const int m = problem.edges.size();

        using hash_map = robin_hood::unordered_map<FrontierState, Int>;

        hash_map nodes;
        nodes[FrontierState{}] = 1;

        hash_map next_nodes;

        // ProgressBar progress_bar(colorlog::success, "Frontier");
        uint64_t zdd_size = 0;
        Int res = 0;
        for (int level = 1; level <= m; ++level) {
            // progress_bar.write(level, m);
            zdd_size += nodes.size();

            std::cerr << "Level " << level << '\n';
            std::cerr << "  Frontier Size: " << FrontierState::frontier_info().frontier.size() << '\n';
            std::cerr << "  Width        : " << nodes.size() << '\n';
            std::cerr.flush();

            FrontierState::set_level(level, info);

            next_nodes.clear();
            // traversing zdd nodes at the level `level`
            for (auto& [prev_state, path_num] : nodes) {
                // 1-edge transition
                if (prev_state.template roughly_check_if_new_edge_can_be_added<check_terminal>()) {
                    FrontierState next_state1 = prev_state;
                    FrontierState::TransitionResult result = next_state1.make_transition<check_terminal>(true, true);
                    if (result == FrontierState::TransitionResult::ONWAY) {
                        next_nodes[std::move(next_state1)] += path_num;
                    } else if (result == FrontierState::TransitionResult::COMPLETE) {
                        res += path_num;
                    }
                }

                // 0-edge transition
                {
                    FrontierState& next_state0 = const_cast<FrontierState&>(prev_state);
                    FrontierState::TransitionResult result = next_state0.make_transition<check_terminal>(false);
                    if (result == FrontierState::TransitionResult::ONWAY) {
                        next_nodes[std::move(next_state0)] += path_num;
                    } else if (result == FrontierState::TransitionResult::COMPLETE) {
                        res += path_num;
                    }
                }
            }
            std::swap(nodes, next_nodes);
        }

        std::cerr << "ZDD Size: " << zdd_size << std::endl;

        return res;
    }

    /**
     * @brief solve instances with a limitation of path length by Simpath algorithm
     * @tparam Int integer type used to count
     * @tparam check_terminal whether checking the conditions about endpoints of paths or not
     * @param problem instance
     * @param edge_ord ordering of edges
     * @return number of paths grouped by length
     */
    template <typename Int, bool check_terminal>
    std::vector<Int> solve_frontier_limited_path(const instance& problem, const std::vector<int>& edge_ord) {
        // count paths by grouping them by length
        // Value is a type for representing pairs of { offset, count }, where count[i] is the number of paths with length offset+i
        using Value = std::pair<uint16_t, std::vector<Int>>;

        const FrontierInfo info(problem, edge_ord);

        FrontierState::set_terminals(problem.s, problem.t);
        FrontierState::init_hasher(problem);

        const int m = problem.edges.size();

        using hash_map = robin_hood::unordered_map<FrontierState, Value>;
        using iterator = typename hash_map::iterator;
        std::vector<iterator> iters;

        hash_map nodes, next_nodes;
        nodes[FrontierState{}] = Value{ 0, std::vector<Int>{ 1 } };

        size_t hit = 0;

        // merge values and store the result to the first argument
        auto merge_value = [&](Value& target, Value&& source){
            if (target.first > source.first) {
                std::swap(target, source);
            }
            const size_t d = source.first - target.first, siz2 = source.second.size();
            if (target.second.size() < d + siz2) {
                target.second.resize(d + siz2);
            }
            for (size_t i = 0; i < siz2; ++i) {
                target.second[d + i] += source.second[i];
            }
        };

        auto insert = [&](hash_map& mp, FrontierState&& state, Value &&path_num) {
            auto [it, inserted] = mp.try_emplace(std::move(state), std::move(path_num));
            if (inserted) return;
            merge_value(it->second, std::move(path_num));
            ++hit;
        };

        // ProgressBar progress_bar(colorlog::success, "Frontier");
        uint64_t zdd_size = 0;
        std::vector<Int> res(problem.l + 2);
        for (int level = 1; level <= m; ++level) {
            hit = 0;

            // progress_bar.write(level, m);
            zdd_size += nodes.size();

            std::cerr << "Level " << level << '\n';
            std::cerr << "  Frontier Size: " << FrontierState::frontier_info().frontier.size() << '\n';
            std::cerr << "  Width        : " << nodes.size() << '\n';
            std::cerr.flush();

            FrontierState::set_level(level, info);

            next_nodes.clear();

            for (auto& [prev_state, path_num] : nodes) {
                // 1-edge transition
                if (prev_state.template roughly_check_if_new_edge_can_be_added<check_terminal>()) {
                    FrontierState next_state1 = prev_state;
                    FrontierState::TransitionResult result = next_state1.make_transition<check_terminal>(true, true);
                    if (result != FrontierState::TransitionResult::FAIL) {
                        // increment the offset
                        ++path_num.first;

                        // number of paths with positive length
                        const int p = next_state1.path_num();
                        // to form a single path, we need to add at least max(p-1,0) edges
                        const int siz = problem.l - path_num.first - p + (p == 0) + 2;
                        // discarding too large sets
                        if (siz > 0) {
                            if (result == FrontierState::TransitionResult::ONWAY) {
                                int nsiz = std::min(siz, int(path_num.second.size()));
                                insert(next_nodes, std::move(next_state1), { path_num.first, std::vector<Int>(path_num.second.begin(), path_num.second.begin() + nsiz) });
                            } else if (result == FrontierState::TransitionResult::COMPLETE) {
                                for (std::size_t i = 0; i < path_num.second.size(); ++i) {
                                    res[path_num.first + i] += path_num.second[i];
                                }
                            }
                        }

                        // restore
                        --path_num.first;
                    }
                }

                // 0-edge transition
                {
                    FrontierState &next_state0 = const_cast<FrontierState&>(prev_state);
                    FrontierState::TransitionResult result = next_state0.make_transition<check_terminal>(false);
                    if (result == FrontierState::TransitionResult::ONWAY) {
                        insert(next_nodes, std::move(next_state0), std::move(path_num));
                    } else if (result == FrontierState::TransitionResult::COMPLETE) {
                        for (std::size_t i = 0; i < path_num.second.size(); ++i) {
                            res[path_num.first + i] += path_num.second[i];
                        }
                    }
                }
            }
            std::swap(nodes, next_nodes);

            std::cerr << "  Hit          : " << hit << std::endl;
        }

        std::cerr << "ZDD Size: " << zdd_size << std::endl;

        res.resize(problem.l + 1);

        return res;
    }
}

template <typename Int>
std::vector<Int> solve_frontier(const instance& problem, const std::vector<int>& edge_ord) {
    if (problem.is_invalid()) {
        return {};
    }
    if (problem.is_trivial()) {
        return { 1 };
    }
    if (problem.l < problem.n - 1) {
        if (problem.s == -1) {
            return detail::solve_frontier_limited_path<Int, false>(problem, edge_ord);
        } else {
            return detail::solve_frontier_limited_path<Int, true>(problem, edge_ord);
        }
    } else {
        if (problem.s == -1) {
            return { detail::solve_frontier<Int, false>(problem, edge_ord) };
        } else {
            return { detail::solve_frontier<Int, true>(problem, edge_ord) };
        }
    }
}
