#pragma once

#include <algorithm>
#include <cassert>
#include <chrono>

#include "instance.hpp"
#include "graph/bfs.hpp"

#include <omp.h>

namespace detail {
    /**
     * @brief solve one-pair instances by naive dfs 
     * @tparam Int integer type used to count
     * @param problem one-pair instance
     * @return number of paths grouped by length
     */
    template <typename Int>
    std::vector<Int> solve_dfs_one_pair(instance problem) {
        const int n = problem.n;

        std::vector<std::vector<int>> adj = problem.adjacent_list();

        std::vector<int> dt = bfs(adj, problem.t);
        for (auto& vs : adj) {
            std::sort(vs.begin(), vs.end(), [&](int u, int v) { return dt[u] < dt[v]; });
        }

        std::vector<int8_t> visited(n);
        visited[problem.s] = true;

        std::vector<Int> ans(problem.l + 1);
        auto dfs = [&](auto dfs, int cur, int len) -> void {
            if (len + dt[cur] > problem.l) return;

            if (cur == problem.t) {
                ++ans[len];
                return;
            }
            for (int nxt : adj[cur]) if (not visited[nxt]) {
                visited[nxt] = true;
                dfs(dfs, nxt, len + 1);
                visited[nxt] = false;
            }
        };
        dfs(dfs, problem.s, 0);
        return ans;
    }

    /**
     * @brief solve all-pairs instances by naive dfs (+ parallelization)
     * @tparam Int integer type used to count
     * @param problem all-pairs instance
     * @return number of paths grouped by length
     */
    template <typename Int>
    std::vector<Int> solve_dfs_all_pairs(const instance& problem) {
        const int n = problem.n;

        const std::vector<std::vector<int>> adj = problem.adjacent_list();

        std::vector<std::vector<Int>> ans_threads(16, std::vector<Int>(problem.l + 1));

#pragma omp parallel for schedule(dynamic,1) num_threads(16)
        for (int s = 0; s < n; ++s) {
            const int tid = omp_get_thread_num();
            std::vector<Int> ans_s(problem.l + 1);

            std::vector<int8_t> visited(n);
            visited[s] = true;

            auto dfs = [&](auto dfs, int cur, int len) -> void {
                if (len > problem.l) return;
                if (s < cur) {
                    ++ans_s[len];
                }

                for (int nxt : adj[cur]) if (not visited[nxt]) {
                    visited[nxt] = true;
                    dfs(dfs, nxt, len + 1);
                    visited[nxt] = false;
                }
            };
            dfs(dfs, s, 0);
            for (int len = 0; len <= problem.l; ++len) {
                ans_threads[tid][len] += ans_s[len];
            }
        }

        std::vector<Int> ans(problem.l + 1);
        for (int i = 0; i < 16; ++i) {
            for (int len = 0; len <= problem.l; ++len) {
                ans[len] += ans_threads[i][len];
            }
        }
        return ans;
    }
}

template <typename Int>
std::vector<Int> solve_dfs(const instance& problem) {
    if (problem.is_invalid()) {
        return std::vector<Int>{};
    }
    if (problem.is_trivial()) {
        return std::vector<Int>{ 1 };
    }
    if (problem.s != -1) {
        return detail::solve_dfs_one_pair<Int>(problem);
    } else {
        return detail::solve_dfs_all_pairs<Int>(problem);
    }
}
