#pragma once

#include <algorithm>
#include <optional>
#include <cassert>
#include <chrono>
#include <iostream>
#include <iomanip>

#include "instance.hpp"
#include "graph/bfs.hpp"

#include <omp.h>

namespace detail {
    /**
     * @brief solve one-pair instances using neighborhood division 
     * @tparam Int integer type used to count
     * @param problem one-pair instance
     * @return number of paths grouped by length
     */
    template <typename Int>
    std::optional<std::vector<Int>> solve_nd_dfs_one_pair(instance problem) {
        std::cerr << "Original: V = " << std::setw(4) << problem.n << ", E = " << std::setw(4) << problem.edges.size() << std::endl;

        auto [new_problem, weight] = problem.reduce_neighborhood_division();
        problem = std::move(new_problem);

        std::cerr << "Reduced : V = " << std::setw(4) << problem.n << ", E = " << std::setw(4) << problem.edges.size() << std::endl;

        std::vector<std::vector<int>> adj = problem.adjacent_list();

        std::vector<int> dt = bfs(adj, problem.t);
        for (auto& vs : adj) {
            std::sort(vs.begin(), vs.end(), [&](int u, int v) { return dt[u] < dt[v]; });
        }

        --weight[problem.s], --weight[problem.t];

        const int half_l = problem.get_l() / 2;

        const int thread_num = std::max(16, omp_get_max_threads());
        std::vector<Int> ans(problem.l + 2);

        std::vector<std::vector<int>> weight_threads(thread_num);
        std::vector<std::vector<Int>> ans_threads(thread_num, std::vector<Int>(problem.l + 2));
        
        auto dfs = [&](auto dfs, int cur, int len, Int w) -> void {
            if (len + dt[cur] > problem.l) return;

            if (len < half_l) {
                // standard dfs
                for (int nxt : adj[cur]) {
                    if (nxt == problem.t) {
                        ans[len + 1] += w;
                    }
                    if (weight[nxt]) {
                        dfs(dfs, nxt, len + 1, w * weight[nxt]--);
                        ++weight[nxt];
                    }
                }
            } else if (len == half_l) {
                const int siz = adj[cur].size();
                // start parallelization
#pragma omp parallel for schedule(dynamic,1) num_threads(16)
                for (int i = 0; i < siz; ++i) {
                    const int tid = omp_get_thread_num();
                    weight_threads[tid] = weight;
                    const int nxt = adj[cur][i];
                    if (nxt == problem.t) {
                        ans_threads[tid][len + 1] += w;
                    }
                    if (weight_threads[tid][nxt]) {
                        dfs(dfs, nxt, len + 1, w * weight_threads[tid][nxt]--);
                        ++weight_threads[tid][nxt];
                    }
                }
            } else {
                // parallelized
                const int tid = omp_get_thread_num();
                for (int nxt : adj[cur]) {
                    if (nxt == problem.t) {
                        ans_threads[tid][len + 1] += w;
                    }
                    if (weight_threads[tid][nxt]) {
                        dfs(dfs, nxt, len + 1, w * weight_threads[tid][nxt]--);
                        ++weight_threads[tid][nxt];
                    }
                }
            }
        };
        dfs(dfs, problem.s, 0, 1);
        ans.resize(problem.l + 1);
        for (int tid = 0; tid < thread_num; ++tid) {
            for (int i = 0; i <= problem.l; ++i) {
                ans[i] += ans_threads[tid][i];
            }
        }
        return ans;
    }

    /**
     * @brief solve all-pairs instances using neighborhood division 
     * @tparam Int integer type used to count
     * @param problem one-pair instance
     * @return number of paths grouped by length
     */
    template <typename Int>
    std::optional<std::vector<Int>> solve_nd_dfs_all_pairs(instance problem) {
        std::cerr << "Original: V = " << std::setw(4) << problem.n << ", E = " << std::setw(4) << problem.edges.size() << std::endl;

        auto [new_problem, weight] = problem.reduce_neighborhood_division();
        problem = std::move(new_problem);

        std::cerr << "Reduced : V = " << std::setw(4) << problem.n << ", E = " << std::setw(4) << problem.edges.size() << std::endl;

        const int n = problem.n;

        const std::vector<std::vector<int>> adj = problem.adjacent_list();

        std::vector<std::vector<Int>> ans_threads(16, std::vector<Int>(problem.l + 1));

        // run dfs starting from each vertex in parallel
#pragma omp parallel for schedule(dynamic,1) num_threads(16)
        for (int s = 0; s < n; ++s) {
            const int tid = omp_get_thread_num();
            std::vector<Int> ans_s(problem.l + 1);

            std::vector<int> weight_s = weight;

            // w: weight of the current path
            auto dfs = [&](auto dfs, int cur, int len, Int w) -> void {
                if (len > problem.l) return;
                if (s < cur) {
                    ans_s[len] += w;
                } else if (s == cur) {
                    // if endpoints of a path are in the same equivalence class, it is counted exactly twice
                    ans_s[len] += w >> 1;
                }

                for (int nxt : adj[cur]) if (weight_s[nxt]) {
                    dfs(dfs, nxt, len + 1, w * weight_s[nxt]--);
                    ++weight_s[nxt];
                }
            };
            dfs(dfs, s, 0, weight_s[s]--);
            for (int len = 0; len <= problem.l; ++len) {
                ans_threads[tid][len] += ans_s[len];
            }
        }

        // collecting results from threads
        std::vector<Int> ans(problem.l + 1);
        for (int i = 0; i < 16; ++i) {
            for (int len = 1; len <= problem.l; ++len) {
                ans[len] += ans_threads[i][len];
            }
        }
        return ans;
    }
}

template <typename Int>
std::optional<std::vector<Int>> solve_nd_dfs(const instance& problem, const int) {
    if (problem.is_invalid()) {
        return std::vector<Int>{};
    }
    if (problem.is_trivial()) {
        return std::vector<Int>{ 1 };
    }
    if (problem.s != -1) {
        return detail::solve_nd_dfs_one_pair<Int>(problem);
    } else {
        return detail::solve_nd_dfs_all_pairs<Int>(problem);
    }
}
