#pragma once

#include <algorithm>
#include <cstdint>
#include <iostream>
#include <vector>

struct unsigned_bigint {
    std::vector<int> digits; // 10-based

    unsigned_bigint() = default;
    unsigned_bigint(__uint128_t v) {
        for (; v; v /= 10) digits.push_back(v % 10);
    }

    friend unsigned_bigint& operator+=(unsigned_bigint &x, const unsigned_bigint &y) {
        if (x.digits.size() < y.digits.size()) {
            x.digits.resize(y.digits.size());
        }
        int carry = 0;
        for (size_t i = 0; i < y.digits.size(); ++i) {
            x[i] += y[i] + carry;
            carry = x[i] >= 10;
            x[i] -= carry * 10;
        }
        for (size_t i = y.digits.size(); carry and i < x.digits.size(); ++i) {
            x[i] += carry;
            carry = x[i] >= 10;
            x[i] -= carry * 10;
        }
        if (carry) x.digits.push_back(1);
        return x;
    }
    friend unsigned_bigint operator+(const unsigned_bigint &x, const unsigned_bigint &y) {
        unsigned_bigint z = x;
        z += y;
        return z;
    }
    friend unsigned_bigint& operator*=(unsigned_bigint &x, const unsigned_bigint &y) {
        return x = x * y;
    }
    friend unsigned_bigint operator*(const unsigned_bigint &x, const unsigned_bigint &y) {
        // naive digit-by-digit multiplication
        if (x.digits.empty() or y.digits.empty()) return 0;
        unsigned_bigint z;
        size_t siz = x.digits.size() + y.digits.size() - 1;
        z.digits.resize(siz);
        for (size_t i = 0; i < x.digits.size(); ++i) for (size_t j = 0; j < y.digits.size(); ++j) {
            z[i + j] += x[i] * y[j];
        }
        int carry = 0;
        for (size_t i = 0; i < siz; ++i) {
            z[i] += carry;
            carry = z[i] / 10;
            z[i] -= carry * 10;
        }
        for (; carry; carry /= 10) {
            z.digits.push_back(carry % 10);
        }
        return z;
    }

    __uint128_t mod(__uint128_t m) const {
        __uint128_t res = 0;
        for (auto it = digits.rbegin(); it != digits.rend(); ++it) {
            __uint128_t nxt = 0;
            for (int i = 0; i < 10; ++i) {
                nxt += res;
                if (nxt >= m) nxt -= m;
            }
            nxt += *it;
            if (nxt >= m) nxt -= m;
            res = nxt;
        }
        return res;
    }

    std::string to_string() const {
        unsigned_bigint tmp = *this;
        tmp.normalize();
        if (tmp.digits.empty()) {
            return "0";
        } else {
            std::string res;
            for (auto it = tmp.digits.rbegin(); it != tmp.digits.rend(); ++it) {
                res += '0' + *it;
            }
            return res;
        }
    }

    friend std::ostream& operator<<(std::ostream& out, const unsigned_bigint& v) {
        return out << v.to_string();
    }
private:
    void normalize() {
        while (digits.size() and digits.back() == 0) digits.pop_back();
    }
    int& operator[](size_t i) {
        return digits[i];
    }
    const int& operator[](size_t i) const {
        return digits[i];
    }
};