import argparse
import os
from pathlib import Path
import signal
import subprocess
import sys
import time
from typing import List, Tuple
from dataclasses import dataclass

from instance import Instance

cmd_meet_in_the_middle = [
    './solve_dfs_meet_in_the_middle.out',
]
cmd_dfs = [
    './solve_nd_dfs.out',
]
cmd_simpath = [
    './solve_frontier.out',
]
cmd_simpath_cluster = [
    './solve_clustering.out',
    '--solver',
    './solve_frontier.out',
]

@dataclass
class Args:
    verbose: bool

def main(args: Args):
    problem = Instance.read_DIMACS(sys.stdin)
    input_text = problem.DIMACS()
    input_data = input_text.encode("utf-8")
    processes : List[Tuple[str, subprocess.Popen]] = []

    def popen(cmd: List[str], err):
        return subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=err,
            start_new_session=True,
        )

    if problem.L <= 14:
        # Meet in the middle
        if args.verbose:
            with Path("log/meet_in_the_middle.log").open(mode='w') as err:
                processes.append(("Meet in the middle", popen(cmd_meet_in_the_middle, err)))
        else:
            processes.append(("Meet in the middle", popen(cmd_meet_in_the_middle, subprocess.DEVNULL)))
    else:
        # Solution 1: DFS
        if args.verbose:
            with Path("log/dfs.log").open(mode='w') as err:
                processes.append(("Naive DFS", popen(cmd_dfs, err)))
        else:
            processes.append(("Naive DFS", popen(cmd_dfs, subprocess.DEVNULL)))
        
        # Solution 2: No Ordering
        if args.verbose:
            with Path("log/frontier.log").open(mode='w') as err:
                processes.append(("Simpath (No Ordering)", popen(cmd_simpath, err)))
        else:
            processes.append(("Simpath (No Ordering)", popen(cmd_simpath, subprocess.DEVNULL)))
        
        # Solution 3: Cluster Ordering
        if args.verbose:
            with Path("log/frontier.log").open(mode='w') as err:
                processes.append(("Simpath (Cluster Ordering)", popen(cmd_simpath_cluster, err)))
        else:
            processes.append(("Simpath (Cluster Ordering)", popen(cmd_simpath_cluster, subprocess.DEVNULL)))
    
    for _, p in processes:
        p.stdin.write(input_data)
        p.stdin.close()
    
    while True:
        time.sleep(1)
        for name, p in processes:
            if p.poll() is not None:
                fastest_solver_name = name
                output = p.stdout.read().decode().strip()
                break
        else:
            continue
        break

    print(output, flush=True)
    if args.verbose:
        print(fastest_solver_name, file=sys.stderr, flush=True)

    for name, p in processes:
        if p.poll() is None:
            os.killpg(os.getpgid(p.pid), signal.SIGTERM)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--verbose', action='store_true')
    args = parser.parse_args()

    verbose = args.verbose
    main(Args(
        verbose=verbose
    ))
