from argparse import ArgumentParser
from pathlib import Path
import subprocess
import sys

from instance import Instance

import networkx as nx
from matplotlib import pyplot as plt
from matplotlib import cm

def get_cmap(n, name='hsv'):
    return cm.get_cmap(name, n + 2)

def reorder_edges_clustering(instance: Instance, draw=False):
    g = nx.Graph()
    g.add_nodes_from(range(1, instance.V + 1))
    g.add_edges_from(instance.E)

    if draw:
        fig, ax = plt.subplots()
        pos = nx.spring_layout(g)
        nx.draw(
            g,
            pos=pos,
            node_color='skyblue',
            edge_color='lightblue',
            with_labels=True,
            ax=ax,
        )
        fig.savefig("figures/graph.png")

    def calc_communities():
        communities = set()

        max_modularity = -1000000000
        communities = set()

        # divide to communities by Girvan-Newman algorithm
        for divisions in nx.algorithms.community.girvan_newman(g):
            modularity = nx.algorithms.community.modularity(g, divisions)
            # adopt a partition with the highest modularity
            if max_modularity < modularity:
                max_modularity = modularity
                communities = divisions
        
        return list(communities)
    
    communities = calc_communities()
    
    if draw:
        step = 0xFFFFFF // len(communities)
        colors = [f'#{hex(i * step)[2:].zfill(6)}' for i in range(len(communities))]

        for id, vs in enumerate(communities):
            for v in vs:
                g.nodes()[v]['color'] = colors[id]
        
        fig, ax = plt.subplots()
        nx.draw(
            g,
            pos=pos,
            node_color=[node[1] for node in g.nodes(data='color')],
            edge_color='black',
            with_labels=True,
            ax=ax
        )
        fig.savefig("figures/cluster.png")

    def calc_vmap():
        """ordering vertices"""
        vmap = {}

        comp_id = dict()
        for i, comp in enumerate(communities):
            for v in comp:
                comp_id[v] = i

        g2 = nx.Graph()
        g2.add_nodes_from(range(len(communities)))
        for u, v in g.edges():
            g2.add_edge(comp_id[u], comp_id[v])
        
        nxt_id = 1

        def sorted_inner(comp):
            """sorting vertices in a community"""
            in_front = lambda x: x in vmap
            in_back = lambda x: x not in vmap and x not in communities[comp]
            keys = { u: (sum(in_back(v) for v in g[u]), -sum(in_front(v) for v in g[u])) for u in communities[comp] }
            return sorted(communities[comp], key=lambda u: keys[u])
        
        def sorted_outer(comps):
            """sorting communities"""
            key = dict()
            for comp in comps:
                S = set(vmap.keys()).union(communities[comp])
                edge_num = 0
                for u, v in g.edges():
                    edge_num += (u in S) and (v not in S)
                    edge_num += (u not in S) and (v in S)
                key[comp] = edge_num
            return sorted(comps, key=lambda comp: key[comp])

        def order_by_bfs():
            """ordering communities based on bfs"""
            source = comp_id[instance.s] if instance.s is not None else 0
            vis = [False] * len(communities)
            vis[source] = True

            que = [source]
            def bfs():
                nonlocal nxt_id

                for cur_comp in que:
                    for v in sorted_inner(cur_comp):
                        vmap[v] = nxt_id
                        nxt_id = nxt_id + 1
                    for nxt_comp in sorted_outer(g2[cur_comp]):
                        if not vis[nxt_comp]:
                            que.append(nxt_comp)
                            vis[nxt_comp] = True
            while True:
                bfs()

                for comp in range(len(communities)):
                    if not vis[comp]:
                        vis[comp] = True
                        que = [comp]
                        break
                else:
                    break

        def order_by_dfs():
            """ordering communities based on dfs"""
            vis = [False] * len(communities)

            def dfs(cur_comp):
                vis[cur_comp] = True

                nonlocal nxt_id
                for v in sorted_inner(cur_comp):
                    vmap[v] = nxt_id
                    nxt_id = nxt_id + 1

                for nxt_comp in sorted_outer(g2[cur_comp]):
                    if vis[nxt_comp]:
                        continue
                    dfs(nxt_comp)

            source = comp_id[instance.s] if instance.s is not None else 0

            while True:
                dfs(source)

                for comp in range(len(communities)):
                    if not vis[comp]:
                        source = comp
                        break
                else:
                    break

        order_by_dfs()
        # order_by_bfs()

        return vmap

    vmap = calc_vmap()
    instance.E = sorted((min(vmap[u], vmap[v]), max(vmap[u], vmap[v])) for u, v in instance.E)

    # vmap = dict()
    # nxt_id = 1
    # source = instance.s if instance.s is not None else 1
    # distances_from_source = nx.single_source_dijkstra_path_length(g, source=source)
    # for vs in sorted(communities, key=lambda comp: sum(distances_from_source[v] for v in comp) / len(comp)):
    #     for v in sorted(vs):
    #         vmap[v] = nxt_id
    #         nxt_id = nxt_id + 1
    # instance.E = sorted((min(vmap[u], vmap[v]), max(vmap[u], vmap[v])) for u, v in instance.E)
    
    if instance.s is not None:
        instance.s = vmap[instance.s]
    if instance.t is not None:
        instance.t = vmap[instance.t]

    if draw:
        h = nx.Graph()
        h.add_nodes_from(range(1, instance.V + 1))
        h.add_edges_from(instance.E)

        for i in range(1, instance.V + 1):
            h.nodes()[vmap[i]]['color'] = g.nodes()[i]['color']
        hpos = { vmap[i]: p for i, p in pos.items() }
        
        fig, ax = plt.subplots()
        nx.draw(
            h,
            pos=hpos,
            node_color=[node[1] for node in h.nodes(data='color')],
            edge_color='black',
            with_labels=True,
            ax=ax
        )
        fig.savefig("figures/ordered.png")


def solve(solver: Path, instance: Instance):
    solver_cmd = [
        f'./{solver}'
    ]
    subprocess.run(
        solver_cmd,
        input=bytes(instance.DIMACS(), encoding='utf-8'),
        check=True
    )

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('--solver', type=str, default='a.out')
    parser.add_argument('--no-order', action='store_true')
    parser.add_argument('--draw', action='store_true')
    args = parser.parse_args()

    solver = Path(args.solver)
    do_ordering = not args.no_order
    draw = args.draw

    instance = Instance.read_DIMACS(sys.stdin)
    if do_ordering:
        print("Begin ordering", file=sys.stderr)
        reorder_edges_clustering(instance, draw=draw)
        print("End ordering", file=sys.stderr)
    solve(solver, instance)