#include <algorithm>
#include <array>
#include <numeric>
#include <iostream>
#include <iomanip>

#include "include/instance.hpp"

#include <chrono>

#include "include/util/io.hpp"

#include "include/util/unsigned_bigint.hpp"
#include "include/util/modint.hpp"

// M1 = 2^128 = 340282366920938463463374607431768211456

// 147808829414345923316083210206383297621 : prime
constexpr __uint128_t M2 = (__uint128_t(0x6f32f1ef8b18a2bcULL) << 64) | 0x3cea59789c79d455ULL;
// 6475935027855235332913700071614325611 = (2^128)^(-1) (mod M2)
constexpr __uint128_t INV_M1 = (__uint128_t(0x4df383f7ddae6e2ULL) << 64) | 0xfe841f8f2962236bULL;

struct Int {
    __uint128_t first;
    modint<M2> second;

    Int() = default;
    Int(__uint128_t v) : first(v), second(v) {}

    friend Int& operator+=(Int& x, const Int& y) {
        x.first += y.first;
        x.second += y.second;
        return x;
    }
    friend Int operator+(const Int& x, const Int& y) {
        Int z = x;
        z += y;
        return z;
    }
};

unsigned_bigint crt(Int x) {
    // ans = q M1 + r
    //      q M1 + r = a (mod M1)
    //      q M1 + r = b (mod M2)
    // 1. r = a
    // 2. q = (b - a) * M1^(-1) (mod M2)

    __uint128_t a = x.first, b = x.second.val;
    __uint128_t b_minus_a = b - a + M2 * (b < a);

    unsigned_bigint r = a;
    unsigned_bigint q = (unsigned_bigint(b_minus_a) * unsigned_bigint(INV_M1)).mod(M2);
    unsigned_bigint res = q * unsigned_bigint(__uint128_t(1) << 64) * unsigned_bigint(__uint128_t(1) << 64) + r;
    return res;
}

#include "include/solve_frontier.hpp"

int main() {
    const Int answer = [] {
        instance problem;
        problem.read_DIMACS(std::cin);
        problem.l = std::min(problem.n - 1, problem.l);

        auto start_time = std::chrono::system_clock::now();

        std::vector<Int> ans = [&] {
            if (problem.s != -1) {
                problem = problem.reduce();
            }
            std::vector<int> edge_ord(problem.edges.size());
            std::iota(edge_ord.begin(), edge_ord.end(), 0);
            return solve_frontier<Int>(problem, edge_ord);
        }();

        Int sum = 0;
        for (Int v : ans) sum += v;
        std::cerr << "Answer = " << crt(sum) << std::endl;
        std::cerr << "Details:" << std::endl;
        for (int i = 0; i < int(ans.size()); ++i) {
            std::cerr << "    Path size " << std::setw(2) << i << ": " << crt(ans[i]) << std::endl;
        }

        auto end_time = std::chrono::system_clock::now();

        auto elapsed_sec = std::chrono::duration_cast<std::chrono::milliseconds>(end_time - start_time).count();

        std::cerr << "Solved Time: " << elapsed_sec << " ms" << std::endl;

        return sum;
    }();

    std::cout << crt(answer) << std::endl;
}