from argparse import ArgumentParser
import os
import signal
from dataclasses import dataclass
import logging
from pathlib import Path
import subprocess
import sys
import time
import psutil
from typing import List, Optional

ROOT = Path(__file__).parent.parent
sys.path.append(str(ROOT / 'src'))

from instance import Instance


num_all = 50
num_one = 100

BENCH_HOME = Path('icgca-benchmarks')
BENCH_ONE = BENCH_HOME / 'one_pair'
BENCH_ALL = BENCH_HOME / 'all_pairs'

ANSWER_ONE = BENCH_HOME / 'out/one_pair'
ANSWER_ALL = BENCH_HOME / 'out/all_pairs'


@dataclass
class Result:
    solved_time_sec: float
    answer: int


def run(cmd : List[str], *, in_file: Path, err_file: Path, ans_file: Path):
    with in_file.open(mode='r') as stdin, err_file.open(mode='w') as stderr:
        logging.info(f'Running {in_file}...')

        start_time_sec = time.time()

        try:
            p = subprocess.Popen(
                cmd,
                stdin=stdin,
                stdout=subprocess.PIPE,
                stderr=stderr,
                start_new_session=True,
            )
        except Exception as e:
            logging.error(e)
            return None

        try:
            output, _ = p.communicate(timeout=600)
        except subprocess.TimeoutExpired as e:
            # 子プロセスもろともkill
            for cp in psutil.Process(p.pid).children(recursive=True):
                cp.kill()
            p.kill()
            # os.killpg(os.getpgid(p.pid), signal.SIGTERM)
            logging.warning('Timeout.')
            return None
        
        end_time_sec = time.time()
        solved_time_sec = (end_time_sec - start_time_sec)

        logging.info(f'Done in {solved_time_sec} seconds.')
        logging.info('Checking output format...')
        try:
            answer, *tails = output.decode().split()
            if len(tails):
                raise Exception('too many output.')
            answer = int(answer)
        except Exception as e:
            logging.error(e)
            return None
        
        logging.info(f'Verifying the answer "{answer}"...')
        try:
            if ans_file.exists():
                with ans_file.open(mode='r') as f:
                    expected_answer, *tails = f.read().split()
                    if len(tails):
                        raise Exception(f'Something wrong with the file "{ans_file}".')
                    expected_answer = int(expected_answer)
                    if answer != expected_answer:
                        raise Exception(f'Wrong Answer.')
                logging.info('Correct.')
            else:
                logging.info(f'The answer is unknown. Writing the answer "{answer}" to "{ans_file}"...')
                ans_file.parent.mkdir(parents=True, exist_ok=True)
                ans_file.touch()
                with ans_file.open(mode='w') as f:
                    f.write(f'{answer}\n')
                logging.info('Done.')
        except Exception as e:
            logging.error(e)
            return None
        
        return Result(
            solved_time_sec=solved_time_sec,
            answer=answer
        )

def run_one_pair(cmd: List[str], err_home: Path, num: int):
    err_one = err_home / 'one_pair'

    for i in range(num, num_one):
        case_name = f'{i:03}.col'
        in_file = BENCH_ONE / case_name
        err_file = err_one / case_name
        ans_file = ANSWER_ONE / case_name

        err_file.parent.mkdir(parents=True, exist_ok=True)

        with in_file.open(mode='r') as f:
            instance = Instance.read_DIMACS(f)

        yield str(in_file.name), instance, run(
            cmd,
            in_file=in_file,
            err_file=err_file,
            ans_file=ans_file
        )

def run_all_pairs(cmd: List[str], err_home: Path, num: int):
    err_all = err_home / 'all_pairs'

    for i in range(num, num_all):
        case_name = f'{i:03}.col'
        in_file = BENCH_ALL / case_name
        err_file = err_all / case_name
        ans_file = ANSWER_ALL / case_name

        err_file.parent.mkdir(parents=True, exist_ok=True)

        with in_file.open(mode='r') as f:
            instance = Instance.read_DIMACS(f)

        yield str(in_file.name), instance, run(
            cmd,
            in_file=in_file,
            err_file=err_file,
            ans_file=ans_file
        )

@dataclass
class Params:
    cmd: List[str]
    err_home: Path
    one_pair: int
    all_pairs: int
    out: Path


@dataclass
class Table:
    name: str
    instance: Instance
    result: Optional[Result]

    @staticmethod
    def header():
        return \
            R'| instance |$\lvert V\rvert$|$\lvert E\rvert$| $l$ | solved time | answer |''\n'\
            R'|---------:|---------------:|---------------:|----:|------------:|-------:|''\n'
    
    def body(self):
        if self.result is None:
            return f'|`{self.name}`|{self.instance.V}|{len(self.instance.E)}|{self.instance.L}| - s| - |\n'
        else:
            return f'|`{self.name}`|{self.instance.V}|{len(self.instance.E)}|{self.instance.L}|{self.result.solved_time_sec:.1f} sec|{self.result.answer}|\n'


def main(params: Params):
    err_home = Path(params.err_home)

    with Path(params.out).open(mode='a') as out:
        out.write('# Result\n\n')

        solved_one_pair = 0
        out.write('## one pair\n\n')
        out.write(Table.header())
        out.flush()
        for name, instance, result in run_one_pair(params.cmd, err_home, params.one_pair):
            solved_one_pair += result is not None
            out.write(Table(name=name, instance=instance, result=result).body())
            out.flush()
        out.write('\n')
        out.write(f'Solved: {solved_one_pair} / {num_one - params.one_pair}\n\n')

        solved_all_pairs = 0
        out.write('## all pairs\n\n')
        out.write(Table.header())
        out.flush()
        for name, instance, result in run_all_pairs(params.cmd, err_home, params.all_pairs):
            solved_all_pairs += result is not None
            out.write(Table(name=name, instance=instance, result=result).body())
            out.flush()
        out.write('\n')
        out.write(f'Solved: {solved_all_pairs} / {num_all - params.all_pairs}\n\n')

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    parser = ArgumentParser()
    parser.add_argument('cmd', type=str)
    parser.add_argument('--out', '-o', type=str)
    parser.add_argument('--all-pairs', type=int, default=0)
    parser.add_argument('--one-pair', type=int, default=0)
    parser.add_argument('--stderr', type=str, default=f'{BENCH_HOME}/stderr/')

    args = parser.parse_args()
    params = Params(
        cmd=args.cmd.split(),
        err_home=Path(args.stderr),
        one_pair=args.one_pair,
        all_pairs=args.all_pairs,
        out=Path(args.out)
    )
    main(params)
